'''
Quarch Power Module Calibration Functions
Written for Python 3.6 64 bit

M Dearman April 2019
'''

'''
Calibration Flow
    Connect to PAM Fixture 
    Connect to Keithley
    step through a set of values and get ADC vs Reference Value
    evaluate results vs defined limits

'''

#Imports QuarchPy library, providing the functions needed to use Quarch modules
from .PowerModuleCalibration import *
from .calibrationConfig import *
from .keithley_2460_control import *
from .QTL2536_6_way_switchbox import *
from quarchpy.device.device import *
from quarchpy.device.scanDevices import userSelectDevice
from quarchpy.user_interface import *
from quarchpy.user_interface import logSimpleResult
from quarchpy.utilities.BitManipulation import *
from .pamHelpers import bcdString

def parseFixtureData(response,start,length):

    # split the multiline response into a list
    response = response.splitlines()
    result = ""
    # for each line
    for line in response:
        # remove 0x, swap bytes
        line = line[4:6] + line[2:4]
        # convert 4 char Hex to 16 bit binary string
        line = "{0:016b}".format(int(line,16))
        # concatenate all the strings
        result += line
    # pick out the section we want
    result = int(result[start:(start+length)],2)
    # convert two's compliment
    if (result >= 2**(length-1)):
        result -= 2**length
    return result

def getFixtureData(device,channel):
    #hold measurement
    response = device.sendCommand("read 0x0000")
    device.sendCommand("write 0x0000 " + setBit(response,3))
    #read measurement
    data = device.sendCommand("read 0x1000 to 0x1008")
    #release measurement
    response = device.sendCommand("read 0x0000")
    device.sendCommand("write 0x0000 " + clearBit(response,3))

    if (channel == "+12V V"):
        return parseFixtureData(data,0,16)
    elif (channel == "+12V dV"):
        return parseFixtureData(data,16,17)
    elif (channel == "+3.3V V"):
        return parseFixtureData(data,33,16)
    elif (channel == "+3.3V dV"):
        return parseFixtureData(data,49,17)
    elif (channel == "+3.3Vaux V"):
        return parseFixtureData(data,66,16)
    elif (channel == "+3.3Vaux dV"):
        return parseFixtureData(data,82,17)
    elif (channel == "+12Vaux V"):
        return parseFixtureData(data,99,16)
    elif (channel == "+12Vaux dV"):
        return parseFixtureData(data,115,17)


class QTL2631_pcie (PowerModule):

    # Fixture Register Addresses
    CAL_ADDRESSES = {
    #'SERIAL_1'						: '0xA102',	-- Mezzanine Serial Number
    #'SERIAL_2'						: '0xA103',	-- Mezzanine Serial Number
    #'SERIAL_3'						: '0xA104',	-- Mezzanine Serial Number
    '+12V_VOLT_MULTIPLIER'       : '0xA105',
    '+12V_VOLT_OFFSET'           : '0xA106',
    '+12V_CURR_MULTIPLIER'       : '0xA107',
    '+12V_CURR_OFFSET'           : '0xA108',
    #'POWER_1_LEAKAGE_MULTIPLIER'   : '0xA109',	-- Not used in External Shunt Mezzanine
    '+3.3V_VOLT_MULTIPLIER'       : '0xA10A',
    '+3.3V_VOLT_OFFSET'           : '0xA10B',
    '+3.3V_CURR_MULTIPLIER'       : '0xA10C',
    '+3.3V_CURR_OFFSET'           : '0xA10D',
    #'+3.3V_LEAKAGE_MULTIPLIER'   : '0xA10E',	-- Not used in External Shunt Mezzanine
    '+3.3Vaux_VOLT_MULTIPLIER'       : '0xA10F',
    '+3.3Vaux_VOLT_OFFSET'           : '0xA110',
    '+3.3Vaux_CURR_MULTIPLIER'       : '0xA111',
    '+3.3Vaux_CURR_OFFSET'           : '0xA112',
    #'+3.3Vaux_LEAKAGE_MULTIPLIER'   : '0xA113',	-- Not used in External Shunt Mezzanine
    '+12Vaux_VOLT_MULTIPLIER'       : '0xA114',
    '+12Vaux_VOLT_OFFSET'           : '0xA115',
    '+12Vaux_CURR_MULTIPLIER'       : '0xA116',
    '+12Vaux_CURR_OFFSET'           : '0xA117',
    'CALIBRATION_COMPLETE'          : '0xA118'
    }

    CAL_CONTROL_SETTINGS = {
    '+12V_CALIBRATION_CONTROL_SETTING'	: '0x00F0',	#set full averaging
    '+3.3V_CALIBRATION_CONTROL_SETTING'	: '0x00F0',	#set full averaging
    '+3.3Vaux_CALIBRATION_CONTROL_SETTING'	: '0x00F0',	#set full averaging
    '+12Vaux_CALIBRATION_CONTROL_SETTING'	: '0x00F0',	#set full averaging
    '+12V_SHUNT'                         : 0.00208,
    '+3.3V_SHUNT'                         : 0.00508,
    '+3.3Vaux_SHUNT'                         : 0.0204,
    '+12Vaux_SHUNT'                         : 0.0004
	}

    CALIBRATION_MODE_ADDR           = '0xA100'
    CALIBRATION_CONTROL_ADDR        = '0xA101'
    LOAD_VOLTAGE					= 12000

    host_switchbox_title = "12V Switchbox"
    host_switchbox_message = "Select the switch box which connects a 12V supply to POWER IN:"
    host_switchbox_mapping = {'12V':'A','+12V_IN':'1','+3.3V_IN':'2','+3.3Vaux_IN':'3','+12Vaux_IN':'4'}

    load_switchbox_title = "Load Switchbox"
    load_switchbox_message = "Select the switch box which connects a Keithley 2460 SourceMeter to POWER OUT:"
    load_switchbox_mapping = {'LOAD':'A','+12V_OUT':'1','+3.3V_OUT':'2','+3.3Vaux_OUT':'3','+12Vaux_OUT':'4'}

    # Fixture Information
    PAMSerial = None
    FixtureSerial = None
    calObjectSerial = None     # The serial number of the device that is being calibrated, i.e QTL1944 in HD PPM, Fixture in PAM
    idnStr = None
    Firmware = None
    Fpga = None
    calInstrument = None
    calInstrumentId = None
    host_switchbox = None
    load_switchbox = None
    waitComplete = False
    checkedWiring = False

    def __init__(self,dut):

        # set the name of this module
        self.name = "Gen5 AIC x16 PAM Fixture with AUX Power"
        self.dut = dut
        
        # Serial numbers (ensure QTL at start)
        self.enclosureSerial = self.dut.sendCommand("*ENCLOSURE?")
        if (self.enclosureSerial.find ("QTL") == -1):
            self.enclosureSerial = "QTL" + self.enclosureSerial
        # fetch the enclosure position
        self.enclosurePosition = self.dut.sendCommand("*POSITION?")
        self.PAMSerial = self.dut.sendCommand ("*SERIAL?")
        if (self.PAMSerial.find ("QTL") == -1):
            self.PAMSerial = "QTL" + self.PAMSerial
        # Fixture Serial
        # fixture serial is retrieved as BCD, we need to convert and pad it
        self.FixtureSerial = "QTL" + bcdString(dut.sendCommand("read 0xA102"),4) + "-" + bcdString(dut.sendCommand("read 0xA103"),2) + "-" + bcdString(dut.sendCommand("read 0xA104"),3) # TODO: this should be replaced with fix:serial? command when implemented
        # calObjectSerial Serial
        self.calObjectSerial = self.FixtureSerial
        # Filename String
        self.filenameString = self.FixtureSerial
        # Code version (FPGA)
        self.idnStr = dut.sendCommand ("*IDN?")
        pos = self.idnStr.upper().find ("FPGA 1:")
        if (pos != -1):
            versionStr = self.idnStr[pos+7:]
            pos = versionStr.find ("\n")
            if (pos != -1):
                versionStr = versionStr[:pos].strip()
            else:
                pass
        else:
            versionStr = "NOT-FOUND"    
        self.Fpga = versionStr.strip()
    
        # Code version (FW)    
        pos = self.idnStr.upper().find ("PROCESSOR:")
        if (pos != -1):
            versionStr = self.idnStr[pos+10:]
            pos = versionStr.find ("\n")
            if (pos != -1):
                versionStr = versionStr[:pos].strip()            
            else:
                pass
        else:
            versionStr = "NOT-FOUND"    
        self.Firmware = versionStr.strip()

        self.calibrations = {}
        # populate +12V channel with calibrations
        self.calibrations["+12V"] = {
            "Voltage":self.QTL2631_pcie_VoltageCalibration(self,"+12V"),
            "DeltaVolts":self.QTL2631_pcie_deltaVoltCalibration(self,"+12V")
            }
        # populate +3.3V channel with calibrations
        self.calibrations["+3.3V"] = {
            "Voltage":self.QTL2631_pcie_VoltageCalibration(self,"+3.3V"),
            "DeltaVolts":self.QTL2631_pcie_deltaVoltCalibration(self,"+3.3V")
            }
        # populate +3.3Vaux channel with calibrations
        self.calibrations["+3.3Vaux"] = {
            "Voltage":self.QTL2631_pcie_VoltageCalibration(self,"+3.3Vaux"),
            "DeltaVolts":self.QTL2631_pcie_deltaVoltCalibration(self,"+3.3Vaux")
            }
        # populate +12Vaux channel with calibrations
        self.calibrations["+12Vaux"] = {
            "Voltage":self.QTL2631_pcie_VoltageCalibration(self,"+12Vaux"),
            "DeltaVolts":self.QTL2631_pcie_deltaVoltCalibration(self,"+12Vaux")
            }

        self.verifications = {}
        # populate +12V channel with verifications
        self.verifications["+12V"] = {
            "Voltage":self.QTL2631_pcie_VoltageVerification(self,"+12V"),
            "Current":self.QTL2631_pcie_deltaVoltVerification(self,"+12V")
            }
        self.verifications["+3.3V"] = {
            "Voltage":self.QTL2631_pcie_VoltageVerification(self,"+3.3V"),
            "Current":self.QTL2631_pcie_deltaVoltVerification(self,"+3.3V")
            }
        self.verifications["+3.3Vaux"] = {
            "Voltage":self.QTL2631_pcie_VoltageVerification(self,"+3.3Vaux"),
            "Current":self.QTL2631_pcie_deltaVoltVerification(self,"+3.3Vaux")
            }
        self.verifications["+12Vaux"] = {
            "Voltage":self.QTL2631_pcie_VoltageVerification(self,"+12Vaux"),
            "Current":self.QTL2631_pcie_deltaVoltVerification(self,"+12Vaux")
            }

    def specific_requirements(self):

        reportText=""

        # select the host switchbox to use for calibration
        if "host_switchbox" in calibrationResources.keys():
            self.host_switchbox = calibrationResources["host_switchbox"]
        else:
            self.host_switchbox = get_switchbox(self.host_switchbox_message,self.host_switchbox_title,self.host_switchbox_mapping)
            calibrationResources["host_switchbox"] = self.host_switchbox

        # select the load switchbox to use for calibration
        if "load_switchbox" in calibrationResources.keys():
            self.load_switchbox = calibrationResources["load_switchbox"]
        else:
            self.load_switchbox = get_switchbox(self.load_switchbox_message,self.load_switchbox_title,self.load_switchbox_mapping)
            calibrationResources["load_switchbox"] = self.load_switchbox
        
        if self.checkedWiring != True:
            self.host_switchbox.checkWiring()
            self.load_switchbox.checkWiring()
            self.checkedWiring = True

        # Select a Keithley SMU
        # If no calibration instrument is provided, request it
        while (True):
            if (calibrationResources["loadString"] == None):
                loadString = userSelectCalInstrument(scanFilterStr="Keithley 2460", nice=True)
                # quit if necessary
                if loadString == 'quit':
                    printText("no module selected, exiting...")
                    sys.exit(0)
                else:
                    calibrationResources["loadString"] = loadString
            try:
                # Connect to the calibration instrument
                self.calInstrument = keithley2460(calibrationResources["loadString"])
                # Open the connection
                self.calInstrument.openConnection()
                self.calInstrumentId = self.calInstrument._sendCommandQuery ("*IDN?")
                break
            # In fail, allow the user to try again with a new selection
            except:
                printText("Unable to communicate with selected instrument!")
                printText("")
                calibrationResources["loadString"] = None

        # Write module specific report header to file
        reportText += "Quarch Power Analysis Module: "
        reportText += self.PAMSerial + "\n"
        reportText += "Quarch Fixture: "
        reportText += self.FixtureSerial + "\n"
        reportText += "Quarch FW Versions: "
        reportText += "FW:" + self.Firmware + ", FPGA: " + self.Fpga + "\n"
        reportText += "\n"
        reportText += "Calibration Instruments#:\n"
        reportText += self.calInstrumentId + "\n"

        # perform uptime check and write to file
        if self.waitComplete != True:
            reportText += self.wait_for_up_time(desired_up_time=600)
            self.waitComplete = True

        return reportText

    def open_module(self):

        # set unit into calibration mode
        self.dut.sendCommand("write " + self.CALIBRATION_MODE_ADDR + " 0xaa55")
        self.dut.sendCommand("write " + self.CALIBRATION_MODE_ADDR + " 0x55aa")

    def clear_calibration(self):

        # set unit into calibration mode
        self.dut.sendCommand("write " + self.CALIBRATION_MODE_ADDR + " 0xaa55")
        self.dut.sendCommand("write " + self.CALIBRATION_MODE_ADDR + " 0x55aa")

        # clear all calibration registers
        for address in self.CAL_ADDRESSES.values():
            self.dut.sendAndVerifyCommand("write " + address + " 0x0000")
        
        # write 0xaa55 to register to calibration complete register to tell module it is calibrated
        self.dut.sendAndVerifyCommand("write " + self.CAL_ADDRESSES["CALIBRATION_COMPLETE"] + " 0xaa55")
        
    def write_calibration(self):

        # write the calibration registers
        # erase the tag memory
        printText("Erasing TAG memory..")
        self.dut.sendCommand("write 0xa200 0x0020")
        # TODO: should check for completion here...
        # wait for 2 seconds for erase to complete
        # check busy
        while checkBit(self.dut.sendCommand("read 0xa200"),8):
            time.sleep(0.1)
        # write the tag memory
        printText("Programming TAG memory...")
        self.dut.sendCommand("write 0xa200 0x0040")        
        # check busy
        while checkBit(self.dut.sendCommand("read 0xa200"),8):
            time.sleep(0.1)

    def close_module(self):

        # reset the fixture FPGA
        self.dut.sendCommand("fixture:reset")

        # close the connection to the calibration instrument
        self.calInstrument.closeConnection()

    def close_all(self):

        #close all attached devices
        self.calInstrument.disable()
        self.calInstrument.closeConnection()
        self.host_switchbox.disable()
        self.host_switchbox.closeConnection()
        self.load_switchbox.disable()
        self.load_switchbox.closeConnection()

    class QTL2631_pcieCalibration (Calibration):

        def __init__(self):
            super().__init__()

        def init_cal(self,voltage):

            # TODO: No Power control at the moment
            # power up
            #self.powerModule.dut.sendAndVerifyCommand("power up")

            # set averaging to max
            self.powerModule.dut.sendAndVerifyCommand("rec:ave 32k")

            #Turn off compression and power up
            self.powerModule.dut.sendAndVerifyCommand("write 0xA010 0x0011")

            # set module into calibration mode (again?)
            self.powerModule.dut.sendCommand("write " + self.powerModule.CALIBRATION_MODE_ADDR + " 0xaa55")   # will not verify
            self.powerModule.dut.sendCommand("write " + self.powerModule.CALIBRATION_MODE_ADDR + " 0x55aa")   # will not verify

            #Reset Keithley
            self.powerModule.calInstrument.reset()

        # check connections to host power and load
        def checkLoadVoltage(self,voltage,tolerance):

            self.powerModule.calInstrument.setReferenceCurrent(0)
            result = self.powerModule.calInstrument.measureLoadVoltage()*1000   # *1000 because we use mV but keithley uses volts
            # check result is in required range
            if (result >= voltage-tolerance) and (result <= voltage+tolerance):
                return True
            else:
                return False

        def finish_cal(self):

            #turn off load
            self.powerModule.calInstrument.disable()

            #turn off switch
            self.powerModule.host_switchbox.setConnections([]) #sendCommand("connect off")
            self.powerModule.load_switchbox.setConnections([]) #sendCommand("connect off")

            # turn dut to autoranging
            self.powerModule.dut.sendAndVerifyCommand("write " + self.powerModule.CALIBRATION_CONTROL_ADDR + " 0x00F0")

        def report(self, action, data):

            report = []

            # send to report file
            report.append("          Pass Level  +/-(" + str(self.absErrorLimit) + str(self.units) + " + " + str(
                self.relErrorLimit) + "%) \n")

            # check errors and generate report
            report.append('\n')

            if action == "calibrate":
                report.append("\t" + '{0:>11}'.format('Reference ') + self.units + '   ' + '{0:>10}'.format(
                    'Raw Value ') + self.units + '   ' + '{0:>10}'.format(
                    'Result ') + self.units + '   ' + '{0:>10}'.format(
                    'Error ') + self.units + '   ' + '{0:>13}'.format(
                    '+/-(Abs Error,% Error)') + ' ' + '{0:>10}'.format('Pass'))
            elif action == "verify":
                report.append("\t" + '{0:>11}'.format('Reference ') + self.units + '   ' + '{0:>10}'.format(
                    'Result ') + self.units + '   ' + '{0:>10}'.format(
                    'Error ') + self.units + '   ' + '{0:>13}'.format(
                    '+/-(Abs Error,% Error)') + '   ' + '{0:>10}'.format('Pass'))

            report.append(
                "==================================================================================================")

            # zero worst case error vars
            worstAbsError = 0
            worstRelError = 0
            worstRef = None
            overallResult = True

            # for each calibration reference
            for thisLine in data:
                reference = thisLine[1]
                ppmValue = thisLine[0]

                # for calibration, replace value with calibrated result
                if action == "calibrate":
                    calibratedValue = self.getResult(ppmValue)
                # otherwise just use ppmValue directly
                else:
                    calibratedValue = ppmValue

                # work out errors
                (actError, errorSign, absError, relError, result) = getError(reference, calibratedValue,
                                                                             self.absErrorLimit, self.relErrorLimit)

                # compare/replace with running worst case
                if absError >= worstAbsError:
                    if relError >= worstRelError:
                        worstAbsError = absError
                        worstRelError = relError
                        worstCase = errorSign + "(" + str(absError) + self.units + "," + "{:.3f}".format(
                            relError) + "%) @ " + '{:.3f}'.format(reference) + self.units

                # update overall pass/fail
                if result != True:
                    overallResult = False

                # generate report
                passfail = lambda x: "Pass" if x else "Fail"
                if action == "calibrate":
                    report.append("\t" + '{:>11.3f}'.format(reference) + '     ' + '{:>10.1f}'.format(
                        ppmValue) + '     ' + '{:>10.1f}'.format(calibratedValue) + '     ' + "{:>10.3f}".format(
                        actError) + '     ' + '{0:>16}'.format(
                        errorSign + "(" + str(absError) + self.units + "," + "{:.3f}".format(
                            relError) + "%)") + '     ' + '{0:>10}'.format(passfail(result)))
                elif action == "verify":
                    report.append("\t" + '{:>11.3f}'.format(reference) + '     ' + '{:>10.1f}'.format(
                        ppmValue) + '     ' + "{:>10.3f}".format(actError) + '     ' + '{0:>16}'.format(
                        errorSign + "(" + str(absError) + self.units + "," + "{:.3f}".format(
                            relError) + "%)") + '     ' + '{0:>10}'.format(passfail(result)))

            report.append(
                "==================================================================================================")
            report.append('\n')

            if action == "calibrate":
                report.append(
                    "Calculated Multiplier: " + str(self.multiplier.originalValue()) + ", Calculated Offset: " + str(
                        self.offset.originalValue()))
                report.append("Stored Multiplier: " + str(self.multiplier.storedValue()) + ", Stored Offset: " + str(
                    self.offset.storedValue()))
                report.append("Multiplier Register: " + self.multiplier.hexString(
                    4) + ", Offset Register: " + self.offset.hexString(4))

            report.append(
                "" + '{0:<35}'.format(self.title) + '     ' + '{0:>10}'.format("Passed : ") + '  ' + '{0:<5}'.format(
                    str(overallResult)) + '     ' + '{0:>11}'.format("worst case:") + '  ' + '{0:>11}'.format(
                    worstCase))
            report.append("\n\n\n")

            # Add to Test Summary? Do this here?
            passfail = lambda x: "Passed" if x else "Failed"

            return {"title": self.title, "result": overallResult, "worst case": worstCase,
                    "report": ('\n'.join(report))}

    class QTL2631_pcie_VoltageCalibration (QTL2631_pcieCalibration):

        def __init__(self,powerModule,thisChannel):

            self.thisChannel = thisChannel
            self.title = thisChannel + " Voltage Calibration"
            self.powerModule = powerModule
            self.absErrorLimit = 2                  # 2mV
            self.relErrorLimit = 1                  # 1%
            self.test_min = 40                      # 40mV
            self.test_max = 14400                   # 14.4V
            self.test_steps = 20
            self.units = "mV"
            self.scaling = 4
            self.multiplier_signed = False
            self.multiplier_int_width = 1
            self.multiplier_frac_width = 16
            self.offset_signed = True
            self.offset_int_width = 10
            self.offset_frac_width = 6
            self.skip_ref_check = True

        def init(self):

            super().init_cal(self.thisChannel)

            # clear the multiplier and offset registers by setting them to zero
            self.powerModule.dut.sendAndVerifyCommand("write " + self.powerModule.CAL_ADDRESSES[self.thisChannel + "_VOLT_MULTIPLIER"] + " 0x0000")
            self.powerModule.dut.sendAndVerifyCommand("write " + self.powerModule.CAL_ADDRESSES[self.thisChannel + "_VOLT_OFFSET"] + " 0x0000")

            self.powerModule.host_switchbox.setConnections([("12V",None)])
            self.powerModule.load_switchbox.setConnections([("LOAD",self.thisChannel + '_OUT')])

            # Setup Keithley
            self.setup()

            # Check Host Power is present
            #while (super().checkLoadVoltage(500,500) != True):
            #    self.powerModule.setConnections("+12V",None,reset=True)

        def setup(self):
            self.powerModule.calInstrument.initialSetupForReferenceVoltage(currentLimit="1e-1")

        def setRef(self,value):

            self.powerModule.calInstrument.setReferenceVoltage(value/1000,currentLimit="1e-1")

        def readRef(self):

            return self.powerModule.calInstrument.measureLoadVoltage()*1000

        def readVal(self):

            return getFixtureData(self.powerModule.dut,self.thisChannel + " V")

        def setCoefficients(self):

            result1 = self.powerModule.dut.sendAndVerifyCommand("write " + self.powerModule.CAL_ADDRESSES[self.thisChannel + "_VOLT_MULTIPLIER"] + " " + self.multiplier.hexString(4))
            result2 = self.powerModule.dut.sendAndVerifyCommand("write " + self.powerModule.CAL_ADDRESSES[self.thisChannel + "_VOLT_OFFSET"] + " " + self.offset.hexString(4))
            if result1 and result2:
                result = True
            else:
                result = False
            logSimpleResult("Set " + self.thisChannel + " voltage", result)

        def finish(self):

            super().finish_cal()

        def report(self,data):

            return super().report("calibrate",data)

        def readCoefficients(self):

            coefficients = {}
            # get voltage multiplier
            coefficients["multiplier"] = self.powerModule.dut.sendCommand("read " + self.powerModule.CAL_ADDRESSES[self.thisChannel + "_VOLT_MULTIPLIER"])
            # get voltage offset
            coefficients["offset"] = self.powerModule.dut.sendCommand("read " + self.powerModule.CAL_ADDRESSES[self.thisChannel + "_VOLT_OFFSET"])
            return coefficients

        def writeCoefficients(self,coefficients):

            # write voltage multiplier
            self.powerModule.dut.sendAndVerifyCommand("write " + self.powerModule.CAL_ADDRESSES[self.thisChannel + "_VOLT_MULTIPLIER"] + " " + coefficients["multiplier"])
            # write voltage offset
            self.powerModule.dut.sendAndVerifyCommand("write " + self.powerModule.CAL_ADDRESSES[self.thisChannel + "_VOLT_OFFSET"] + " " + coefficients["offset"])

    class QTL2631_pcie_deltaVoltCalibration (QTL2631_pcieCalibration):

        def __init__(self,powerModule,thisChannel):

            self.thisChannel = thisChannel
            self.title = thisChannel + " deltaVolt Calibration"
            self.powerModule = powerModule
            self.absErrorLimit = 10                 # 10uV
            self.relErrorLimit = 1                  # 1%
            self.test_min = 10                      # 10uV
            self.test_max = 50000                   # 50000uV
            if self.test_max > (self.powerModule.CAL_CONTROL_SETTINGS[self.thisChannel + "_SHUNT"]*4)*1000000:  # if test_max is going to push the Keithely over 4 amps
                self.test_max = (self.powerModule.CAL_CONTROL_SETTINGS[self.thisChannel + "_SHUNT"]*4)*1000000    # limit test max to whatever the keithley can source
            self.test_steps = 20
            self.units = "uV"
            self.scaling = 32
            self.multiplier_signed = False
            self.multiplier_int_width = 1
            self.multiplier_frac_width = 16
            self.offset_signed = True
            self.offset_int_width = 10
            self.offset_frac_width = 6
            self.skip_ref_check = True

        def init(self):

            super().init_cal(self.thisChannel)

            #set manual range, full averaging, [thisChannel] current range
            self.powerModule.dut.sendAndVerifyCommand("write " + self.powerModule.CALIBRATION_CONTROL_ADDR + " " + self.powerModule.CAL_CONTROL_SETTINGS[self.thisChannel + "_CALIBRATION_CONTROL_SETTING"])
            # clear the multiplier and offset registers by setting them to zero
            self.powerModule.dut.sendAndVerifyCommand("write " + self.powerModule.CAL_ADDRESSES[self.thisChannel + "_CURR_MULTIPLIER"] + " 0x0000")
            self.powerModule.dut.sendAndVerifyCommand("write " + self.powerModule.CAL_ADDRESSES[self.thisChannel + "_CURR_OFFSET"] + " 0x0000")

            self.powerModule.host_switchbox.setConnections([("12V",self.thisChannel + '_IN')])
            self.powerModule.load_switchbox.setConnections([("LOAD",self.thisChannel + '_OUT')])

            # Setup Keithley
            self.setup()

            # Check Host Power is present
            while (super().checkLoadVoltage(self.powerModule.LOAD_VOLTAGE,1000) != True):
                showDialog(message="Host Power not detected, please check the connections",title="Check Connections")
                self.powerModule.host_switchbox.setConnections([("12V",self.thisChannel + '_IN')],reset=True)
                self.powerModule.load_switchbox.setConnections([("LOAD",self.thisChannel + '_OUT')],reset=True)

        def setup(self):
            self.powerModule.calInstrument.initialSetupForReferenceCurrent()

        def setRef(self,value):

            # The Keithley sources current in Amps, the calibration board has a 5mOhm resistor on it
            # we need to convert our reference voltage into an amp value
            # i = V/R = (value in uV/1000000) / Resistance
            self.powerModule.calInstrument.setReferenceCurrent((value/1000000)/self.powerModule.CAL_CONTROL_SETTINGS[self.thisChannel + "_SHUNT"])

        def readRef(self):

            # The Keithley sources resistance in Amps, the calibration board has a 5mOhm resistor on it
            # we need to convert the amp value back into a voltage in uV
            # V= IR = current*resistance*1000000           
            return self.powerModule.calInstrument.measureLoadCurrent()*self.powerModule.CAL_CONTROL_SETTINGS[self.thisChannel + "_SHUNT"]*1000000

        def readVal(self):

            return getFixtureData(self.powerModule.dut,self.thisChannel + " dV")

        def setCoefficients(self):

            result1 = self.powerModule.dut.sendAndVerifyCommand("write " + self.powerModule.CAL_ADDRESSES[self.thisChannel + "_CURR_MULTIPLIER"] + " " + self.multiplier.hexString(4))
            result2 = self.powerModule.dut.sendAndVerifyCommand("write " + self.powerModule.CAL_ADDRESSES[self.thisChannel + "_CURR_OFFSET"] + " " + self.offset.hexString(4))
            if result1 and result2:
                result = True
            else:
                result = False
            logSimpleResult("Set " + self.thisChannel + " current", result)

        def finish(self):

            super().finish_cal()

        def report(self,data):

            return super().report("calibrate",data)

        def readCoefficients(self):

            coefficients = {}
            # get current multiplier
            coefficients["multiplier"] = self.powerModule.dut.sendCommand("read " + self.powerModule.CAL_ADDRESSES[self.thisChannel + "_CURR_MULTIPLIER"])
            # get current offset
            coefficients["offset"] = self.powerModule.dut.sendCommand("read " + self.powerModule.CAL_ADDRESSES[self.thisChannel + "_CURR_OFFSET"])
            return coefficients

        def writeCoefficients(self,coefficients):

            # write current multiplier
            self.powerModule.dut.sendAndVerifyCommand("write " + self.powerModule.CAL_ADDRESSES[self.thisChannel + "_CURR_MULTIPLIER"] + " " + coefficients["multiplier"])
            # write current offset
            self.powerModule.dut.sendAndVerifyCommand("write " + self.powerModule.CAL_ADDRESSES[self.thisChannel + "_CURR_OFFSET"] + " " + coefficients["offset"])

    class QTL2631_pcie_VoltageVerification (QTL2631_pcieCalibration):

        def __init__(self,powerModule,thisChannel):

            self.thisChannel = thisChannel
            self.title = thisChannel + " Voltage Verification"
            self.powerModule = powerModule
            self.absErrorLimit = 2      # 2mV
            self.relErrorLimit = 1      # 1%
            self.test_min = 40          # 40mV
            self.test_max = 14400       # 14.4V
            self.test_steps = 20
            self.units = "mV"
            self.skip_ref_check = True

        def init(self):

            super().init_cal(self.thisChannel)

            self.powerModule.host_switchbox.setConnections([("12V",None)])
            self.powerModule.load_switchbox.setConnections([("LOAD",self.thisChannel + '_OUT')])

            # Setup Keithley
            self.setup()

            # Check Host Power is present
            #while (super().checkLoadVoltage(500,500) != True):
                #self.powerModule.setConnections(self.thisChannel,None,reset=True)gma

        def setup(self):
            self.powerModule.calInstrument.initialSetupForReferenceVoltage(currentLimit="1e-1")

        def setRef(self,value):

            self.powerModule.calInstrument.setReferenceVoltage(value/1000,currentLimit="1e-1")

        def readRef(self):

            return self.powerModule.calInstrument.measureLoadVoltage()*1000

        def readVal(self):

            return getFixtureData(self.powerModule.dut,self.thisChannel + " V")

        def finish(self):

            super().finish_cal()

        def report(self,data):

            return super().report("verify",data)

    class QTL2631_pcie_deltaVoltVerification (QTL2631_pcieCalibration):

        def __init__(self,powerModule,thisChannel):

            self.thisChannel = thisChannel
            self.title = thisChannel + " deltaVolt Verification"
            self.powerModule = powerModule
            self.absErrorLimit = 10     # 10uV = 2mA  - Deliberately higher than calibration limit
            self.relErrorLimit = 1      # 1% tolerance
            self.test_min = 10          # 10uV
            self.test_max = 50000       # 50000uV
            if self.test_max > (self.powerModule.CAL_CONTROL_SETTINGS[self.thisChannel + "_SHUNT"]*4)*1000000:  # if test_max is going to push the Keithely over 4 amps
                self.test_max = (self.powerModule.CAL_CONTROL_SETTINGS[self.thisChannel + "_SHUNT"]*4)*1000000    # limit test max to whatever the keithley can source
            self.test_steps = 20
            self.units = "uV"
            self.skip_ref_check = True

        def init(self):

            super().init_cal(self.thisChannel)

            self.powerModule.host_switchbox.setConnections([("12V",self.thisChannel + '_IN')])
            self.powerModule.load_switchbox.setConnections([("LOAD",self.thisChannel + '_OUT')])

            # Setup Keithley
            self.setup()

            # Check Host Power is present
            while (super().checkLoadVoltage(self.powerModule.LOAD_VOLTAGE,1000) != True):
                showDialog(message="Host Power not detected, please check the connections",title="Check Connections")
                self.powerModule.host_switchbox.setConnections([("12V",self.thisChannel + '_IN')],reset=True)
                self.powerModule.load_switchbox.setConnections([("LOAD",self.thisChannel + '_OUT')],reset=True)

        def setup(self):

            self.powerModule.calInstrument.initialSetupForReferenceCurrent()

        def setRef(self,value):

            # The Keithley sources resistance in Amps, the calibration board has a 5mOhm resistor on it
            # we need to convert our reference voltage into an amp value
            # i = V/R = (value in uV/1000000) / Resistance
            self.powerModule.calInstrument.setReferenceCurrent((value/1000000)/self.powerModule.CAL_CONTROL_SETTINGS[self.thisChannel + "_SHUNT"])

        def readRef(self):

            # The Keithley sources resistance in Amps, the calibration board has a 5mOhm resistor on it
            # we need to convert the amp value back into a voltage in uV
            # V= IR = current*resistance*1000000           
            return self.powerModule.calInstrument.measureLoadCurrent()*self.powerModule.CAL_CONTROL_SETTINGS[self.thisChannel + "_SHUNT"]*1000000

        def readVal(self):

            return getFixtureData(self.powerModule.dut,self.thisChannel + " dV")

        def finish(self):

            super().finish_cal()

        def report(self,data):

            return super().report("verify",data)